import logging
import numpy as np
from os.path import join, basename, splitext
from subprocess import call

import pyBigWig

from joblib import Parallel, delayed


def create_bigwigs(matrix, outdir, args):
    """Create bigwigs from matrix."""
    call("mkdir -p {}".format(outdir), shell=True)
    genome_size_dict = args.chromosome_sizes

    outpaths, data = [], []
    for bed_file in matrix:
        outpath = join(outdir, splitext(basename(bed_file))[0] + ".bw")
        outpaths.append(outpath)

        bed_column = matrix[bed_file]
        data.append(bed_column)

    Parallel(n_jobs=args.number_cores)(delayed(_create_bigwig)(bed_column, outpath, genome_size_dict) for outpath, bed_column in zip(outpaths, data))


def _to_int(l):
    return [int(i) for i in l]


def _create_bigwig(bed_column, outpath, genome_size_dict):

    logging.info("Creating biwgwig " + outpath)

    rpkm = 1e6 * bed_column / bed_column.sum()

    rpkm = [float(f) for f in list(rpkm.fillna(0).values.flatten())]

    bed_column = bed_column.reset_index()
    unique_chromosomes = list(bed_column.Chromosome.drop_duplicates())
    chromosomes = list(bed_column.Chromosome)
    starts = _to_int(list(bed_column.Bin))
    ends = _to_int(list(bed_column.End))

    header = [(c, int(genome_size_dict[c])) for c in unique_chromosomes]

    bw = pyBigWig.open(outpath, "w")
    bw.addHeader(header)

    bw.addEntries(chromosomes, starts, ends=ends, values=rpkm)
    bw.close()


def create_sum_bigwigs(matrix, outdir, args):
    call("mkdir -p {}".format(outdir), shell=True)

    chip = matrix[args.treatment].sum(axis=1)
    input = matrix[args.control].sum(axis=1)

    chip_outpath = join(outdir, "chip_sum" + ".bw")
    input_outpath = join(outdir, "input_sum" + ".bw")

    # _create_bigwig(chip, chip_outpath, args)
    # _create_bigwig(input, input_outpath, args)

    Parallel(n_jobs=args.number_cores)(delayed(_create_bigwig)(bed_column, outpath, args.chromosome_sizes) for outpath, bed_column in zip([chip_outpath, input_outpath], [chip, input]))



import pytest

from joblib import Parallel, delayed
from collections import defaultdict
import pandas as pd
import numpy as np

import pkg_resources, os
from natsort import natsorted

from io import StringIO

# from helper.functions

import logging
from rpy2.robjects import r, pandas2ri
pandas2ri.activate()

from rpy2.robjects.packages import importr
importr("S4Vectors")
bioc = importr("GenomicRanges")

length_of_rle = r("function (x) sum(runLength(x))")


__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"


def nucleotide_overlaps_per_file(all_files, nb_cpu):

    rles = files_to_chromosome_coverage(all_files, nb_cpu)

    nucleotide_overlaps = Parallel(n_jobs=nb_cpu)(delayed(_nucleotide_overlaps_per_file)(
        f, rles) for f in rles)

    return pd.concat(nucleotide_overlaps).sort_values(["Main", "Other"]).reset_index(drop=True)


def _nucleotide_overlaps_per_file(bed_file, extended_rles):

    base_bed = bed_file.split("/")[-1].split(".")[0]
    logging.info("Finding the number of nucleotides in " + base_bed + " overlapping other files.")

    _rle_to_iranges = r("""function(x, c){
    gr = GRanges(c, IRanges(start(x), width=runLength(x)), value=runValue(x))
    gr[elementMetadata(gr)$value==1]
    }""")

    _find_overlaps = r("function(x, o) subjectHits(findOverlaps(subject=x, query=o))")
    _set_metadata = r("function(x, ov, f) {elementMetadata(x) = 0; elementMetadata(x[ov]) = 1; data.frame(Start=start(x), End=end(x), Data=elementMetadata(x))}")

    cvs = extended_rles[bed_file]
    irs = {c: _rle_to_iranges(d, c) for c, d in cvs.items()}

    overlaps = defaultdict(list)
    for f in extended_rles:
        base_bed_other = f.split("/")[-1].split(".")[0]
        cvos = extended_rles[f]
        iros = {c: _rle_to_iranges(d, c) for c, d in cvos.items()}

        overlapping_chromosomes = set(irs.keys()).intersection(iros.keys())
        for c in overlapping_chromosomes:
            ir, iro = irs[c], iros[c]
            fov = _find_overlaps(ir, iro)
            df = _set_metadata(ir, fov, base_bed_other)
            df = pandas2ri.ri2py(df)
            df.insert(2, "Chromosome", c)
            df = df.set_index("Start End Chromosome".split())
            df.columns = [base_bed_other]
            overlaps[c].append(df)

    rowdicts = []
    for c, v in natsorted(overlaps.items()):
        df = pd.concat(v, axis=1).reset_index()
        lengths = df.End - df.Start
        df = df.set_index("Chromosome Start End".split())
        for f in df:
            # print(f)
            idx = df.loc[df[f] != 0, f]
            total_overlaps = lengths.ix[idx].sum()
            rowdict = {"Chromosome": c, "Main": base_bed, "Other": f, "Overlaps": total_overlaps}
            rowdicts.append(rowdict)

    return pd.DataFrame.from_dict(rowdicts)["Chromosome                    Main                    Other  Overlaps".split()]


def overlap_matrix_nucleotides(all_files, nb_cpu):
    rles = files_to_chromosome_coverage(all_files, nb_cpu)

    nucleotide_overlaps = Parallel(n_jobs=nb_cpu)(delayed(_overlap_matrix_nucleotides)(
        f, rles) for f in rles)

    df = pd.concat(nucleotide_overlaps)

    return df.reset_index(drop=True)


def _overlap_matrix_nucleotides(bed_file, extended_rles):

    overlaps = _create_overlap_matrix_nucleotides(bed_file, extended_rles)
    return _counts_runlengths(bed_file, overlaps)


def files_to_chromosome_coverage(all_files, nb_cpu):

    df_to_coverage = r("function(x) coverage(GRanges(x$Chromosome, IRanges(x$Start, x$End)))")

    logging.info("Finding nucleotide coverage of files.")
    coverages = defaultdict(dict)
    for f in all_files:
        df = pd.read_table(f, usecols=[0, 1, 2], header=None, names="Chromosome Start End".split())
        cv = df_to_coverage(df)

        chromosomes = r["names"](cv)
        for chromosome in chromosomes:
            coverages[f][chromosome] = r("function(x, idx) x[[idx]]")(cv, chromosome)

    max_per_chromosome_coverage = r("function(x) sum(runLength(x))")
    remove_duplicate_list_entries = r("function(x) x[unique(names(x))]")

    maxlengths = defaultdict(int)
    for f, data in coverages.items():
        for chromosome, rle in data.items():
            current_len = max_per_chromosome_coverage(rle)[0]
            maxlengths[chromosome] = max(maxlengths[chromosome], current_len)

    extend_rle = r('function(cvg, maxlen) c(cvg,Rle(0,maxlen-length(cvg)))')
    extended_rles = defaultdict(dict)
    for f, d in coverages.items():
        for chromosome, cv in d.items():
            maxlength = maxlengths[chromosome]
            cv = extend_rle(cv, maxlength)
            extended_rles[f][chromosome] = cv

    return extended_rles



def _create_overlap_matrix_nucleotides(bed_file, coverages):

    base_bed_other = bed_file.split("/")[-1]
    logging.info("Processing {} at nucleotide level".format(base_bed_other))

    cvs = coverages[bed_file]

    files = [f for f in coverages if f != bed_file]
    compute_overlap = r("function(x, o) x + (x & o)")
    for f in files:
        # print(f)
        other_cvs = coverages[f]
        overlapping_chromosomes = set(cvs.keys()).intersection(other_cvs.keys())

        for chromosome in overlapping_chromosomes:
            chr_cov = cvs[chromosome]
            other_cov = other_cvs[chromosome]

            overlap = compute_overlap(chr_cov, other_cov)
            cvs[chromosome] = overlap

    return cvs


def _counts_runlengths(bed_file, cvs):

    base_bed = bed_file.split("/")[-1].split(".")[0]

    overlaps = defaultdict(int)
    get_runlength = r("function(x, v) runLength(x[x == v])")
    for chromosome, overlap in cvs.items():
        run_values = set(r["runValue"](overlap))
        for run_value in run_values:
            run_value = int(run_value)
            runlength = get_runlength(overlap, run_value)[0]
            overlaps[run_value] += runlength

    rowdicts = []
    for run_value, run_length in overlaps.items():
        rowdict = {"Main": base_bed, "Other": run_value, "Overlaps": run_length}
        rowdicts.append(rowdict)

    return pd.DataFrame.from_dict(rowdicts)


    # command = "bedtools makewindows -w {} -b {} | bedtools intersect -wo -filenames -a - -b {}".format(window_size, bed_file, all_files_str)
    # output = check_output(command, shell=True).decode()

    # df = pd.read_table(StringIO(output), usecols=[0, 1, 2, 3], header=None, names="Chromosome Start End File".split())

    # df.File = df.File.apply(basename).str.split(".", expand=True).ix[:, 0]

    # df.insert(0, "Main", base_bed_other)

    # return _compute_nucleotide_overlap(df, window_size)



# def make_treatment_control_same_length(treatment, control):

#     treatment_result_dict = dict()
#     control_result_dict = dict()
#     for chromosome in treatment:

#         if chromosome not in control:
#             logging.info("Chromosome " + chromosome +
#                          " missing from input data. Skipping.")
#             continue

#         chr_treatment, chr_control = treatment[chromosome], control[chromosome]

#         treatment_maxlen = r["max"](r["sapply"](list(chr_treatment.values()),
#                                                 length_of_rle))
#         control_maxlen = r["max"](r["sapply"](list(chr_control.values()),
#                                               length_of_rle))

#         maxlen = r["max"](treatment_maxlen, control_maxlen)

#         lapply = r('function(cvg, maxlen) c(cvg,Rle(0,maxlen-length(cvg)))')



import logging
from os.path import dirname, join, basename
from subprocess import call

import pandas as pd

from joblib import Parallel, delayed

from natsort import natsorted

from epic.windows.count.remove_out_of_bounds_bins import remove_bins_with_ends_out_of_bounds
from epic.config.genomes import get_genome_size_file



def write_matrix_files(chip_merged, input_merged, df, args):

    matrixes = create_matrixes(chip_merged, input_merged, df, args)

    if args.store_matrix:
        print_matrixes(matrixes, args)

    matrix = pd.concat(matrixes, axis=0)
    matrix = matrix.drop("Enriched", axis=1)
    ends = matrix.index.get_level_values("Bin") + int(args.window_size) - 1
    matrix.insert(0, "End", ends)
    matrix = matrix.set_index("End", append=True)
    matrix = matrix.sort_index(level="Chromosome")

    # TODO: remove out of bounds bins

    if args.bigwig:
        # defer initialization so not run during travis
        from epic.bigwig.create_bigwigs import create_bigwigs
        create_bigwigs(matrix, args.bigwig, args)

    if args.sum_bigwig:
        # defer initialization so not run during travis
        from epic.bigwig.create_bigwigs import create_sum_bigwigs
        create_sum_bigwigs(matrix, args.sum_bigwig, args)


def _create_matrixes(chromosome, chip, input, islands, chromosome_size,
                     window_size):

    chip_df = get_chromosome_df(chromosome, chip)
    input_df = get_chromosome_df(chromosome, input)

    chip_df["Chromosome"] = chip_df["Chromosome"].astype("category")
    chip_df["Bin"] = chip_df["Bin"].astype(int)
    chip_df = chip_df.set_index("Chromosome Bin".split())
    chip_df = islands.join(chip_df, how="right")
    chip_df = chip_df[~chip_df.index.duplicated(keep='first')]

    input_df["Chromosome"] = input_df["Chromosome"].astype("category")
    input_df["Bin"] = input_df["Bin"].astype(int)
    input_df = input_df.set_index("Chromosome Bin".split())
    input_df = input_df[~input_df.index.duplicated(keep='first')]

    dfm = chip_df.join(input_df, how="outer", sort=False).fillna(0)

    dfm = remove_bins_with_ends_out_of_bounds(dfm, chromosome_size,
                                              window_size)

    return dfm


def create_matrixes(chip, input, df, args):

    "Creates matrixes which can be written to file as is (matrix) or as bedGraph."

    genome = args.chromosome_sizes

    chip = put_dfs_in_chromosome_dict(chip)
    input = put_dfs_in_chromosome_dict(input)
    all_chromosomes = natsorted(set(list(chip.keys()) + list(input.keys())))

    islands = enriched_bins(df, args)

    logging.info("Creating matrixes from count data.")
    dfms = Parallel(n_jobs=args.number_cores)(delayed(_create_matrixes)(
        chromosome, chip, input, islands, genome[chromosome],
        args.window_size) for chromosome in all_chromosomes)

    return dfms


def print_matrixes(matrixes, args):

    outpath = args.store_matrix

    dir = dirname(outpath)
    if dir:
        call("mkdir -p {}".format(dir), shell=True)

    logging.info("Writing data matrix to file: " + outpath)
    for i, df in enumerate(matrixes):

        if i == 0:
            header, mode = True, "w+"
        else:
            header, mode = False, "a"

        df.astype(int).to_csv(outpath,
                              sep=" ",
                              na_rep="NA",
                              header=header,
                              mode=mode,
                              compression="gzip",
                              chunksize=1e6)


def get_island_bins(df, window_size, genome):
    """Finds the enriched bins in a df."""

    # need these chromos because the df might not have islands in all chromos
    chromosomes = natsorted(list(args.chromosome_sizes))

    chromosome_island_bins = {}
    df_copy = df.reset_index(drop=False)
    for chromosome in chromosomes:
        cdf = df_copy.loc[df_copy.Chromosome == chromosome]
        if cdf.empty:
            chromosome_island_bins[chromosome] = set()
        else:
            island_starts_ends = zip(cdf.Start.values.tolist(),
                                     cdf.End.values.tolist())
            island_bins = chain(*[range(
                int(start), int(end), window_size)
                                  for start, end in island_starts_ends])
            chromosome_island_bins[chromosome] = set(island_bins)

    return chromosome_island_bins


def put_dfs_in_dict(dfs):

    sample_dict = {}
    for df in dfs:

        if df.empty:
            continue

        chromosome = df.head(1).Chromosome.values[0]
        sample_dict[chromosome] = df

    return sample_dict


def put_dfs_in_chromosome_dict(dfs):

    chromosome_dict = {}
    for df in dfs:

        if df.empty:
            continue

        chromosome = df.head(1).Chromosome.values[0]
        chromosome_dict[chromosome] = df

    return chromosome_dict


def get_chromosome_df(chromosome, df_dict):

    if chromosome in df_dict:
        df = df_dict[chromosome]
    else:
        df = pd.DataFrame(columns="Chromosome Bin".split())

    return df


def enriched_bins(df, args):

    df = df.loc[df.FDR < args.false_discovery_rate_cutoff]

    idx_rowdicts = []
    for _, row in df.iterrows():
        for bin in range(
                int(row.Start), int(row.End) + 2, int(args.window_size)):
            idx_rowdicts.append({"Chromosome": row.Chromosome,
                                 "Bin": bin,
                                 "Enriched": 1})
    islands = pd.DataFrame.from_dict(idx_rowdicts)
    islands.loc[:, "Chromosome"].astype("category")
    islands.loc[:, "Bin"].astype(int)

    return islands.set_index("Chromosome Bin".split())
